// CVCC Power Regulator - Front Panel Board Firmware / for ATtiny2313
// 2013.6.23 Eleken
// rev_c1815@yahoo.co.jp
// http://www1.adachi.ne.jp/2sc1815/electronics/

// main clock : 1.0 MHz
// Fuse bits  : ext  0b11111111 (same as default)
//            : high 0b11011011 (enables BOD reset)
//            : low  0b01100100 (same as default)

#include "main.h"

/* Grobals */

// current digit
volatile uint8_t g_digit[8];
 
// current switch state (0xFF : unpushed, otherwise : see header file)
volatile uint8_t g_switchState; 

const uint8_t g_fontTable[10] = {
0b00111111, 0b00000110, 0b01011011, 0b01001111, 0b01100110, 
0b01101101, 0b01111101, 0b00100111, 0b01111111, 0b01101111}; // '0' - '9'


/* Timer */
static void InitTimer(void);

// output data to sreg 74xx164
static void OutputSREG(uint8_t data){
	uint8_t i;
	for(i=0; i<8; i++){
		if(data & 0x80) SREG_PORT |= _BV(SREG_DATA);
		else SREG_PORT &= ~_BV(SREG_DATA);
		SREG_PORT |= _BV(SREG_CLK);
		SREG_PORT &= ~_BV(SREG_CLK);
		
		data = data << 1;
	}
}

// output data to sreg 74xx164
// with pin shuffling
static void OutputSREGEx(uint8_t data){
	uint8_t i;
	const uint8_t pinlist[8] = {
		0b00010000, 0b00001000, 0b00000100, 0b10000000,
		0b01000000, 0b00000010, 0b00000001, 0b00100000};
	for(i=0; i<8; i++){
		if(data & pinlist[i]) SREG_PORT |= _BV(SREG_DATA);
		else SREG_PORT &= ~_BV(SREG_DATA);
		SREG_PORT |= _BV(SREG_CLK);
		SREG_PORT &= ~_BV(SREG_CLK);
	}
}


// timer1 : about 1 ms per interrupt
ISR(TIMER1_COMPA_vect){
	static uint8_t state = 0;
	
	if(state & 1){ 
		if(!(state & 8)){ // ON cycle 1
			OCR1A = TIMER_ON1;
		}else{
			OCR1A = TIMER_ON2;
		}
	}else if(state == 0){
		OCR1A = 255;
	}else{// OFF cycle 
		OCR1A = TIMER_OFF;
	}
	
	uint8_t seg = 0; // 

	seg = g_digit[(state>>1)];
	// reset common pins
	PORTB = (PORTB & 0b11100000) | 0b00001111;
	PORTD = (PORTD & 0b00001111) | 0b00000000;
	
	if(g_digit[5] & 0x80) SBI(PORTB, PB6);
	else CBI(PORTB, PB6);

	// pin output
	if(state & 8){
		OutputSREGEx(~seg);
	}else{
		OutputSREGEx(seg);
	}
	
	if(state & 1){
		switch (state>>1){
		case 0: // 
			SBI(PORTB, PB4);
			break;
		case 1: // 
			SBI(PORTD, PD6);
			break;
		case 2: // 
			SBI(PORTD, PD5);
			break;
		case 3: // 
			SBI(PORTD, PD4);
			break;

		case 4: // 
			CBI(PORTB, PB3);
			break;
		case 5: // 
			CBI(PORTB, PB2);
			break;
		case 6: // 
			CBI(PORTB, PB1);
			break;
		default:
			CBI(PORTB, PB0);
			break;
		}
	}
	
	if(state == 0){ // check switch matrix (about per 15ms)
		static uint8_t cnt;
		if(cnt++ >= 4){
			cnt = 0;
			static uint8_t sw_old;
			static uint8_t longpush_cnt;
			uint8_t sw = 0xFF;
			
			uint8_t temp;
			OutputSREGEx(0xFF);
			// Row 1
			PORTB = (PORTB & 0b11100000) | 0b00000111;
			NOP();
			temp = PIND & 0x0F;
			if(temp == 0b00000111) sw = SW_7;
			else if(temp == 0b00001011) sw = SW_8;
			else if(temp == 0b00001101) sw = SW_9;
			else if(temp == 0b00001110) sw = SW_CV;
			else if(temp == 0b00000110) sw = SW_SETCC;
			else if(temp != 0b00001111) sw = SW_UNKNOWN; 
			
			// Row 2
			PORTB = (PORTB & 0b11100000) | 0b00001011;
			NOP();
			temp = PIND & 0x0F;
			if(temp == 0b00000111) sw = SW_4;
			else if(temp == 0b00001011) sw = SW_5;
			else if(temp == 0b00001101) sw = SW_6;
			else if(temp == 0b00001110) sw = SW_CC;
			else if(temp == 0b00000110) sw = SW_DISCONNECT;
			else if(temp != 0b00001111) sw = SW_UNKNOWN;
			
			// Row 3
			PORTB = (PORTB & 0b11100000) | 0b00001101;
			NOP();
			temp = PIND & 0x0F;
			if(temp == 0b00000111) sw = SW_1;
			else if(temp == 0b00001011) sw = SW_2;
			else if(temp == 0b00001101) sw = SW_3;
			else if(temp == 0b00001110) sw = SW_CANCEL;
			else if(temp != 0b00001111) sw = SW_UNKNOWN;
			
			// Row 4
			PORTB = (PORTB & 0b11100000) | 0b00001110;
			NOP();
			temp = PIND & 0x0F;
			if(temp == 0b00000111) sw = SW_0;
			else if(temp == 0b00001011) sw = SW_CLEAR;
			else if(temp == 0b00001101) sw = SW_DOT;
			else if(temp == 0b00001110) sw = SW_ENTER;
			else if(temp != 0b00001111) sw = SW_UNKNOWN;
			
			if((sw != 0xFF) && (longpush_cnt > LONGPUSH_WAIT)){ // being long-pushed
				g_switchState = SW_PUSHING | sw;
			}else if((sw_old != sw) && (sw == 0xFF)){ // switch state has changed, turned to off
				if(longpush_cnt == 0) { // too short
				}else if(longpush_cnt <= LONGPUSH_WAIT) {// < 500ms
					g_switchState = sw_old; // short push
				}else {
					g_switchState = SW_LONG | sw_old; // long push
				}
			}
			
			if(sw != 0xFF){ // Pushed
				longpush_cnt ++;
			}else{
				longpush_cnt = 0;
			}
			sw_old = sw;
			
			// Clear
			PORTB = (PORTB & 0b11100000) | 0b00001111;
			// reset timer
			TCNT1 = 0;
		}
	}
	
	state++;
	if(state >= 16) state = 0;
}

// USI 
static void InitUsi(void);
static void UsiComplete(void);

// 2-WIRE DATA FORMAT
// DATA LENGTH : 10(SEND x 1 + RECEIVE x 9) 

volatile uint8_t g_ovfState = 0;
volatile uint8_t g_rcvCnt = 0;
volatile uint8_t g_rcvBuffer[USI_BUFFER_SIZE];

ISR(USI_START_vect){ // detected start condition
	g_ovfState = 0;
	CBI(DDRB, PB5); // set SDA to input
	
//	_delay_ms(1);
	
	while((PINB & _BV(PB7)) && !(PINB & _BV(PB5)));
	
	USIDR = 0xFF;
	USICR = 0b11111000; // start condition interrupt & overflow interrupt enable
	USISR = 0b11100000; // conuter = 0, release SCL
	g_rcvCnt = 0;
}

ISR(USI_OVERFLOW_vect){
	// note that this interrupt will occur on falling edge of SCL
//	_delay_ms(1);
	switch(g_ovfState){
	case 0: // after receive address
		if(USIDR == (I2C_ADDR<<1)){
			g_ovfState = 11;
			
			// send ACK
			USIDR = 0x00; // send ACK
			SBI(DDRB, PB5); // set SDA to output
			USICR = 0b11111000; // enable SCL hold by overflow
			USISR = 0b11101110; // counter = 14, release SCL
		}else{
			USIDR = 0xFF; // don't send ACK
			CBI(DDRB, PB5); // set SDA to input
			USICR = 0b10101000; // disable overflow interrupt and SCL hold by overflow
			USISR = 0b11100000; // counter = 0, release SCL
		}
		break;
	case 4: // after receive data
		g_rcvBuffer[g_rcvCnt++] = USIDR;
	case 12: // after send data
		USIDR = 0xFF;
//	_delay_ms(1);
		if(g_rcvCnt >= USI_BUFFER_SIZE) {
			g_ovfState = 21;
		}else {
			g_ovfState = 1;
		}
		// send ACK
		USIDR = 0x00;
		SBI(DDRB, PB5); // set SDA to output
		USICR = 0b11111000; // enable SCL hold by overflow 
		USISR = 0b11101110; // counter = 14, release SCL
		break;
	case 1: // after send ACK
		USIDR = 0xFF;
		CBI(DDRB, PB5); // set SDA to input
		g_ovfState = 4;
		USISR = 0b11100000; // counter = 0, release SCL
		break;
	case 11: // after send FIRST BYTE ACK
		USIDR = g_switchState;
		g_switchState = SW_UNPUSHED;
		SBI(DDRB, PB5); // set SDA to output
		g_ovfState = 12;
		USISR = 0b11100000; // counter = 0, release SCL
		break;
	case 21: // after send LAST BYTE ACK
		USIDR = 0xFF;
		CBI(DDRB, PB5); // set SDA to input
		USICR = 0b10101000; // overflow interrupt disable, SCL hold by overflow disabve
		USISR = 0b11100000; // release SCL
		g_ovfState = 255;
		UsiComplete();
		break;
	default: // error?
		InitUsi();
		break;
	}
}
static void InitUsi(void){
	// 2-wire mode
	SBI(PORTB, PB5); // set SDA to H
	SBI(PORTB, PB7); // set SCL to H
	CBI(DDRB, PB5); // set SDA to input
	SBI(DDRB, PB7); // set SCL to output
	
	USIDR = 0xFF;
	USICR = 0b10101000; // start condition interrupt enable, 2-wire mode
	USISR = 0b11100000; // clear interrupt flag, set counter to zero
}

static void UsiComplete(void){
	for(uint8_t i=0; i<8; i++){
		g_digit[i] = g_rcvBuffer[i];
	}
}

static void InitTimer(void){
	// timer1 (1 ms)
	TCCR1A = 0b00000000; // CTC
	TCCR1B = 0b00001010; // clk/8
	OCR1A  = 125; // ... for 1.0Mhz clock
	TIMSK = 0b01000000; // Compare match interrupt enable
	TCNT1 = 0;
}
// Initialize
static uint8_t Init(void){
	g_switchState = SW_UNPUSHED;

	cli();
	// init clock
//	CLKPR = 0x80;
//	CLKPR = 0x00; // CLKPS = 0 (8MHz)
	
	ACSR = 0b10000000;
	
	
	// init pins
	PORTA = 0b00000000;
	DDRA  = 0b00000011;
	
	PORTB = 0b10101111;
	DDRB  = 0b11011111;
	
	PORTD = 0b00001111;
	DDRD  = 0b01110000;
	
	
	// Check serial IF mode
//	g_mode = PINB & 0b01000000;	// init timer
	InitTimer();

	// USI
	InitUsi();
	
	for(uint8_t i=0; i<8; i++) {
		g_digit[i] = 0x40;
	}
	
	sei();
	
	return 0;
	
}

int main(void){
	
	// 
	if(Init()){
		return -1;
	}
	
	if(0){ // Test bench
		// init timer
		InitTimer();
		uint8_t temp = 0;
		uint8_t i;
		while(1){
			temp++;
			for(i=0; i<50; i++) {
				_delay_ms(10);
			}
			
			for(i=0; i<8; i++) {
				g_digit[i] = g_fontTable[(temp+i) % 10] | 0x80;
			}
			if(g_switchState != SW_UNPUSHED){
				for(i=0; i<6; i++) {
					g_digit[i] = 0;
				}
				uint8_t test = g_switchState;
				g_switchState = SW_UNPUSHED;
				g_digit[6] = g_fontTable[test/10];
				g_digit[7] = g_fontTable[test%10];
			}
		}
	}else{ // normal operation
		while(1){
		}
	}

	return 0;
}
